// EpicsProcessVariable.h: interface for the CEpicsProcessVariable class.
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EPICSPROCESSVARIABLE_H__A2EA2EC6_E7A5_11D2_8F1D_00105AC8D47C__INCLUDED_)
#define AFX_EPICSPROCESSVARIABLE_H__A2EA2EC6_E7A5_11D2_8F1D_00105AC8D47C__INCLUDED_

// EpicsProcessVariable handles the CA stuff:
// Holds the chid, queries DBR_CTRL_*, subscribes to native value.
//
// Abstarct GenericValue and its derived DoubleValue etc.
// handle the different data types.
//
// ProcessVariable (the COM interface) registers callbacks.

#include <cadef.h>

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class CEpicsProcessVariable;

class CGenericValue
{
public:
	CGenericValue ()
	{
		m_units[0] = '\0';
		m_count = 0;
	}

	virtual ~CGenericValue ()
	{
		m_count = 0;
	}

	virtual BOOL Init (CEpicsProcessVariable *epv, unsigned long count)
	{
		m_epv = epv;
		return TRUE;
	}

	// The DBR_xx used for reading this PV:
	virtual chtype GetRequestType () = 0;

	// The DBR_CTRL_ - type for this PV:
	virtual chtype GetControlType ()
	{
		// maybe there is an 'official' way of
		// getting the CTRL-type for a given type ?
		return GetRequestType () + DBR_CTRL_STRING - DBR_STRING;
	}

	// called as a response to GetControlType()
	virtual void CtrlCallback (struct event_handler_args *args) = 0;

	// called repeatedly for monitored PVs
	virtual void EventCallback (struct event_handler_args *args) = 0;

	// Return a VARIANT for this type
	// The VARIANT will be initialized
	// and has to be free'd with VariantClear ()
	virtual BOOL GetValue (VARIANT *dest) = 0;

	char m_units[MAX_UNITS_SIZE];
	unsigned long m_count;
protected:
	CEpicsProcessVariable	*m_epv;
};

class CDoubleValue : public CGenericValue
{
public:
	CDoubleValue ();

	virtual ~CDoubleValue ();

	BOOL Init (CEpicsProcessVariable *epv, unsigned long count);

	chtype GetRequestType ();

	void CtrlCallback (struct event_handler_args *args);

	void EventCallback (struct event_handler_args *args);

	// Return a VARIANT for this type
	// The VARIANT will be initialized
	// and has to be free'd with VariantClear ()
	BOOL GetValue (VARIANT *dest);

private:
	double	*m_value;
	short	m_precision;
};

class CI16Value : public CGenericValue
{
public:
	CI16Value ();

	virtual ~CI16Value ();

	BOOL Init (CEpicsProcessVariable *epv, unsigned long count);

	chtype GetRequestType ();

	void CtrlCallback (struct event_handler_args *args);

	void EventCallback (struct event_handler_args *args);

	// Return a VARIANT for this type
	// The VARIANT will be initialized
	// and has to be free'd with VariantClear ()
	BOOL GetValue (VARIANT *dest);

private:
	short	*m_value;
};

// This is uncomplete!!
class CStringValue : public CGenericValue
{
public:
	CStringValue ()
	{
		m_value[0] = '\0';
	}

	BOOL Init (CEpicsProcessVariable *epv, unsigned long count)
	{
    	return CGenericValue::Init (epv, count);
    }

	chtype GetRequestType ()
	{
		return DBR_STRING;
	}

	void CtrlCallback (struct event_handler_args *args)
	{
		struct dbr_sts_string *data = (struct dbr_sts_string *) args->dbr;

		strcpy (m_value, data->value);
	}

	void EventCallback (struct event_handler_args *args)
	{
		dbr_string_t *data = (dbr_string_t *) args->dbr;

		strcpy (m_value, *data);
		ATLTRACE ("CStringValue::EventCallback: got string '%s'\n", m_value);
	}

	BOOL GetValue (VARIANT *dest);

private:
	dbr_string_t	m_value;
};


typedef void (*EPVNotifyFunc) (void *arg);

class CEpicsProcessVariable  
{
public:
	/**
	 * No public constructor,
	 * use Create instead.
	 *
	 * The supplied notify routine will be called upon
	 * successful connection.
	 *
	 * Returns NULL on error.
	 */
	static CEpicsProcessVariable *Create(const char *name,	EPVNotifyFunc connected, void *arg);

	virtual ~CEpicsProcessVariable();

	/**
	 * Has to be called periodically to keep CA communication alive.
	 */
    static void CAPoll()	{ ca_poll(); }
	const char *GetName()   { return m_name; }
	BOOL IsConnected()      { return m_connected; }
	short GetNativeType()   { return ca_get_field_type(m_chid); }

	const char *GetUnits()  { return IsConnected() ? m_value->m_units : "<unknown>"; }

	// The VARIANT will be initialized
	// and has to be free'd with VariantClear ()
	BOOL GetValue(VARIANT *dest)
    { return (m_connected && m_value) ? m_value->GetValue(dest) : FALSE; }

	BOOL SetShortValue(const short *value, unsigned long count);
	BOOL SetLongValue(const long *value, unsigned long count);
	BOOL SetFloatValue (const float *value, unsigned long count);
	BOOL SetDoubleValue (const double *value, unsigned long count);
	BOOL SetStringValue (const char *value);

    // Estabilish subscription.
    // func will be called for each new value.
    // While subscribed, IssueGetCallback can be called to
    // ask for a new value round-trip. Same func will be called.
	void StartMonitor(EPVNotifyFunc func, void *arg);
    void IssueGetCallback();
	void StopMonitor();

    // One-shot: Use callback for next 'put'
    void UsePutCallback()   { m_use_put_callback = TRUE; }
    void NoPutCallback()    { m_use_put_callback = FALSE; }

private:
	CEpicsProcessVariable();
	// Create a CGenericValue-derived handler for the given native type
	CGenericValue *CreateValue (chtype type, unsigned long count);

	static void ConnectCallback (struct connection_handler_args args);
	static void CtrlCallback(struct event_handler_args args);
	static void EventCallback (struct event_handler_args args);
	static void PutCallback (struct event_handler_args args);

	char			*m_name;
	chid			m_chid;
	evid			m_evid;
	BOOL			m_connected;
	CGenericValue	*m_value;
    BOOL            m_use_put_callback;

	EPVNotifyFunc	m_connection_func;
	void			*m_connection_func_arg;
	EPVNotifyFunc	m_new_value_func;
	void			*m_new_value_func_arg;
};

#endif // !defined(AFX_EPICSPROCESSVARIABLE_H__A2EA2EC6_E7A5_11D2_8F1D_00105AC8D47C__INCLUDED_)
