// ProcessVariable.cpp : Implementation of CProcessVariable
#include "stdafx.h"
#include "Config.h"
#include "EpicsCAClient.h"
#include "ProcessVariable.h"

#ifdef _DEBUG
#include "VariantTools.h"
#endif

/////////////////////////////////////////////////////////////////////////////
// CProcessVariable
//
// This is our exported ActiveX coClass.
// Is will transfer most calls to its
// CEpicsProcessVariable member.

#ifdef _DEBUG
SLList<CProcessVariable> debug_PV_list;
#endif

CProcessVariable::CProcessVariable ()
{
	m_epv = 0;
#ifdef _DEBUG
	debug_PV_list.add (this);
#endif
}

CProcessVariable::~CProcessVariable ()
{
#ifdef _DEBUG
	debug_PV_list.remove (this);
#endif
	if (m_epv)
    {
		delete m_epv;
        m_epv = 0;
    }
}

STDMETHODIMP CProcessVariable::InterfaceSupportsErrorInfo(REFIID riid)
{
	return InlineIsEqualGUID(IID_IProcessVariable, riid) ? S_OK : S_FALSE;
}

// Setting the name property will
// start the process of setting up a CA connection,
// getting control information etc.
//
// At the end an IsConnected-event will (hopefully) be generated.
STDMETHODIMP CProcessVariable::put_name(BSTR newVal)
{
	USES_CONVERSION;
	const char *name = OLE2CA(newVal);

	if (m_epv)
	{
		Error ("Once set, name of PV cannot be changed", IID_IProcessVariable);
		return E_FAIL;
	}

	if (strlen(name) < 1)
	{
		Error ("PV name cannot be empty", IID_IProcessVariable);
		return E_FAIL;
	}

	m_epv = CEpicsProcessVariable::Create (name, ConnectCallback, this);
	if (! m_epv)
	{
		Error ("Cannot create PV (internal error)", IID_IProcessVariable);
		return E_FAIL;
	}

	return S_OK;
}

// Called by CEpicsProcessVariable when connection is OK:
void CProcessVariable::ConnectCallback (void *arg)
{
	CProcessVariable *me = (CProcessVariable *) arg;

	if (! me)
		return;

	if (me->m_epv && me->m_epv->IsConnected())
	{
		me->m_epv->StartMonitor (NewValueCallback, me);
	}
	else
	{
		// Show disconnect (no real value)
		VARIANT	dummy;
		VariantInit (&dummy);
		me->Fire_NewValue (0, dummy);
	}
}

STDMETHODIMP CProcessVariable::get_name(BSTR *pVal)
{
	USES_CONVERSION;
	if (pVal)
		*pVal = m_epv ? A2BSTR (m_epv->GetName ()) :
						A2BSTR ("<uninitialized>");

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_units(BSTR *pVal)
{
	USES_CONVERSION;
	if (pVal)
		*pVal = m_epv ? A2BSTR (m_epv->GetUnits ()) :
						A2BSTR ("<uninitialized>");
	return S_OK;
}

// Called by CEpicsProcessVariable Monitor for new values:
void CProcessVariable::NewValueCallback(void *arg)
{
	CProcessVariable *me = (CProcessVariable *) arg;

	if (me && me->m_epv)
	{
		VARIANT	value;
    
#if TRACELEVEL >= 3
		ATLTRACE("CProcessVariable::NewValueCallback(%s): ", me->m_epv->GetName ());
#endif
		if (me->m_epv->GetValue (&value))
		{
#if TRACELEVEL >= 3
    		ShowVariant (value);
#endif
			me->Fire_NewValue (1, value);
			VariantClear (&value);
		}
	}
}

STDMETHODIMP CProcessVariable::get_value(VARIANT *pVal)
{
	if (m_epv && m_epv->GetValue (pVal))
		return S_OK;

	Error ("PV is not connected", IID_IProcessVariable);
	return E_FAIL;
}

// Called with a VARIANT for the new value.
// Unclear: Am I supposed to call VariantClear(new_value)?
// This might be a memory leak for arrays....
STDMETHODIMP CProcessVariable::put_value(VARIANT new_value)
{
	USES_CONVERSION;
	VARTYPE type = new_value.vt & VT_TYPEMASK;

	if (m_epv == NULL)
	{
		Error ("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}

#if TRACELEVEL >= 4
	ATLTRACE ("CProcessVariable::PutValue (%s)\n", m_epv->GetName());
	ShowVariant (new_value);
#endif

	// handle arrays:
	if (new_value.vt & VT_ARRAY)
	{
		UINT dim = SafeArrayGetDim(new_value.parray);
		long lower, upper;
		unsigned long size;
		void HUGEP *data;

		if (dim != 1)
		{
			Error ("Can only handle single-dimensional arrays", IID_IProcessVariable);
			return E_FAIL;
		}

		//	'1' for first dimension, not '0' !
		SafeArrayGetLBound(new_value.parray, 1, &lower);
		SafeArrayGetUBound(new_value.parray, 1, &upper);
		size = upper - lower + 1; // lower..upper -> 0...size-1

		SafeArrayAccessData (new_value.parray, &data);

		switch (type)
		{
		case VT_I2:
			m_epv->SetShortValue ((short *)data, size);
			break;
		case VT_I4:
			m_epv->SetLongValue ((long *)data, size);
			break;
		case VT_R4:
			m_epv->SetFloatValue ((float *)data, size);
			break;
		case VT_R8:
			m_epv->SetDoubleValue ((double *)data, size);
			break;
		default:
			Error ("Can only handle arrays of short,int,long,float,double. Cannot handle VARIANT arrays of VARIANTS.", IID_IProcessVariable);
			SafeArrayUnaccessData (new_value.parray);
			return E_FAIL;
		}
		SafeArrayUnaccessData (new_value.parray);
		return S_OK;
	}

	// Dispatch VARIANT type.
	switch (type)
	{
	case VT_I2:
		if (m_epv->SetShortValue (&new_value.iVal, 1))
			return S_OK;
		break;
	case VT_I4:
		if (m_epv->SetLongValue (&new_value.lVal, 1))
			return S_OK;
		break;
	case VT_R4:
		if (m_epv->SetFloatValue (&new_value.fltVal, 1))
			return S_OK;
		break;
	case VT_R8:
		if (m_epv->SetDoubleValue (&new_value.dblVal, 1))
			return S_OK;
		break;
	default:
		// Don't know what type was provided,
		// but we can still try to convert it into a string:
		VARIANT converted;
		VariantInit (&converted);
		if (VariantChangeType (&converted, &new_value, 0, VT_BSTR) == S_OK &&
			m_epv->SetStringValue (OLE2A(converted.bstrVal)))
		{
			VariantClear (&converted);
			return S_OK;
		}
	}

	Error ("Can only handle: short,int,long,float,double,string", IID_IProcessVariable);
	return E_FAIL;
}

STDMETHODIMP CProcessVariable::get_type(short *pVal)
{
	USES_CONVERSION;
	if (pVal && m_epv)
	{
		*pVal = m_epv->GetNativeType();
		return S_OK;
	}
	Error ("PV not connected", IID_IProcessVariable);
	return E_FAIL;
}

STDMETHODIMP CProcessVariable::get_is_connected(BOOL *pVal)
{
	*pVal = m_epv->IsConnected ();
	return S_OK;
}

STDMETHODIMP CProcessVariable::put_callback(VARIANT newVal)
{
	if (m_epv == NULL)
	{
		Error ("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}

#if TRACELEVEL >= 4
	ATLTRACE ("CProcessVariable::put_callback(%s)\n", m_epv->GetName());
	ShowVariant (newVal);
#endif

    m_epv->UsePutCallback();
    HRESULT result = put_value(newVal);
    m_epv->NoPutCallback();

	return result;
}

STDMETHODIMP CProcessVariable::get_callback()
{
    m_epv->IssueGetCallback();
	return S_OK;
}
