// EpicsPV.cpp: implementation of the CEpicsPV class.
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "EpicsCAServer.h"
#include "Config.h"
#include "EpicsPV.h"
#include "Server.h"
#include "osiTime.h"
#include "ProcessVariable.h"
#include "float.h"

// Epics Epoch as FILETIME.
// See StaticInit & UpdateValue for details:
static LARGE_INTEGER	EpicsEpoch;

// static members
CServer *CEpicsPV::m_server = NULL;
gddAppFuncTable<CEpicsPV> CEpicsPV::m_func_table;
aitUint32 CEpicsPV::m_value_app;

// Static initialization,
// automatically called when CServer is created:
void CEpicsPV::StaticInit (CServer *server)
{
	if (! m_server) // init. only once
	{
		SYSTEMTIME	st_epoch;
		FILETIME	ft_epoch;

		// Epics Epoch: Jan 1, 1990 (Monday)
		st_epoch.wYear = 1990;
		st_epoch.wMonth = 1;
		st_epoch.wDayOfWeek = 1;
		st_epoch.wDay = 1; 
		st_epoch.wHour = 0;
		st_epoch.wMinute = 0;
		st_epoch.wSecond = 0;
		st_epoch.wMilliseconds = 0; 

		SystemTimeToFileTime (&st_epoch, &ft_epoch);
		EpicsEpoch.LowPart = ft_epoch.dwLowDateTime;
		EpicsEpoch.HighPart = ft_epoch.dwHighDateTime;

		//	Read-routine for value:
		CEpicsPV::m_func_table.installReadFunc ("value", CEpicsPV::getValue);
		//	These are needed in addition to satisfy DBR_GR_xx:
		CEpicsPV::m_func_table.installReadFunc ("units", CEpicsPV::getUnits);
		CEpicsPV::m_func_table.installReadFunc ("alarmLowWarning", CEpicsPV::getAlarmLowWarning);
		CEpicsPV::m_func_table.installReadFunc ("alarmHighWarning", CEpicsPV::getAlarmHighWarning);
		CEpicsPV::m_func_table.installReadFunc ("alarmLow", CEpicsPV::getAlarmLow);
		CEpicsPV::m_func_table.installReadFunc ("alarmHigh", CEpicsPV::getAlarmHigh);
		CEpicsPV::m_func_table.installReadFunc ("graphicLow", CEpicsPV::getGraphicLow);
		CEpicsPV::m_func_table.installReadFunc ("graphicHigh", CEpicsPV::getGraphicHigh);
		CEpicsPV::m_func_table.installReadFunc ("precision", CEpicsPV::getPrecision);
		CEpicsPV::m_func_table.installReadFunc ("controlLow", CEpicsPV::getControlLow);
		CEpicsPV::m_func_table.installReadFunc ("controlHigh", CEpicsPV::getControlHigh);
		CEpicsPV::m_func_table.installReadFunc ("enums", CEpicsPV::getEnums);

		m_value_app =
			gddApplicationTypeTable::app_table.getApplicationType(GDD_NAME_VALUE);
	}
	m_server = server;
}

// constructor cannot return status,
// so this Create-routine should be used to obtain
// a new object.
// Create also handles registration with server:
CEpicsPV *CEpicsPV::Create(const char *name, CProcessVariable *pv)
{
	CEpicsPV *epv = new CEpicsPV (name, pv);
	return epv;
}

CEpicsPV::CEpicsPV(const char *name, CProcessVariable *pv) : stringId (name)
{
	m_pv = pv;

	m_keep_registered = TRUE;
	m_value = NULL;
	m_have_previous_value = FALSE;

	m_units = "a.u.";

	m_deadband = 0.0;
	m_changed = TRUE;

	m_alarm_low = -DBL_MAX;
	m_alarm_low_warning = -DBL_MAX;
	m_alarm_high_warning = DBL_MAX;
	m_alarm_high = DBL_MAX;
	m_graphic_low = -10.0;
	m_graphic_high = +10.0;
	m_precision = 2;
	m_control_low = -10.0;
	m_control_high = +10.0;

	m_monitored = FALSE;

	if (m_server)
		m_server->Register (*this);
}

CEpicsPV::~CEpicsPV ()
{
	if (m_value)
		m_value->unreference ();
#ifdef _DEBUG
	m_value = NULL;
#endif
}

// Nomen est omen:
const char *CEpicsPV::getName() const
{
	// return inherited portion of EpicsPV, stringId:
	return ((stringId *) this)->resourceName();
}

// Associate an enum-string with index.
// From now on this EpicsPV tries to behave as an aitEnum16
BOOL CEpicsPV::SetEnumString (int index, const char *text)
{
	if (index < 0)
		return FALSE;

	if (m_enum_strings.size() <= index)
		m_enum_strings.resize(index+1);
	m_enum_strings[index] = text;

#if TRACELEVEL >= 3
	ATLTRACE ("PV '%s': Enum String #%d = '%s'\n",
		getName(), index, text);
	for (int i=0; i<m_enum_strings.size(); ++i)
		ATLTRACE ("\tstate %d: '%s'\n",
			i, m_enum_strings[i].c_str());
#endif
	
	return m_enum_strings.size() > index;
}

// Called by the server libary each time that it wishes to
// subscribe for PV change notification from the server 
// tool via postEvent() below.
caStatus CEpicsPV::interestRegister ()
{
	m_monitored = TRUE;
	return S_cas_success;
}

// called by the server library each time that it wishes to
// remove its subscription for PV value change events
// from the server tool via caServerPostEvents()
void CEpicsPV::interestDelete ()
{
	m_monitored = FALSE;
}

// What should be used to retrieve the native type?
aitEnum CEpicsPV::bestExternalType() const
{
	// We have enum strings?
	if (m_enum_strings.size() > 0)
		return aitEnumEnum16;

	// Default if value's type not known, yet
	if (! m_value)
		return aitEnumFloat64;
	return m_value->primitiveType();
}

// dimension (0: scalar, 1: array)
unsigned CEpicsPV::maxDimension() const
{
	return m_value ? m_value->dimension () : 0;
}

// return array size for given dimension (starting with dimension=0)
aitIndex CEpicsPV::maxBound (unsigned dimension) const
{
	aitIndex start, nelem;

	if (dimension != 0)
		return 0;

	if (m_value && m_value->dimension()>0)
	{
			m_value->getBound(0, start, nelem);
			return nelem;
	}

	return 1;
}

/*static*/ void CEpicsPV::getEpicsTime(struct timespec &now)
{
	// As long as it's not part of osiTime for certain:
	// Fetch system clock to incorporate adjustments by SoftTimeSlave
	// software:
	FILETIME ft_now;
	LARGE_INTEGER li_now;

	// Get current time (not local time but GMT!)
	GetSystemTimeAsFileTime(&ft_now);
	li_now.LowPart  = ft_now.dwLowDateTime;
	li_now.HighPart = ft_now.dwHighDateTime;
	li_now.QuadPart -= EpicsEpoch.QuadPart;	// Convert to Epics Epoch

	// extract secs/ nanosecs from 100-nanosecond:
	DWORD secpart = DWORD(li_now.QuadPart / 10000000L);
	DWORD hundredNsecpart = DWORD(li_now.QuadPart - UInt32x32To64(secpart,10000000L));
	now.tv_sec  = secpart;
	now.tv_nsec = hundredNsecpart * 100;
}

// Called after the "value" of m_value is set
// to update the time stamp, status and severity
void CEpicsPV::UpdateValue(aitUint16 stat, aitUint16 sevr, struct timespec &time)
{
    m_value->setTimeStamp(&time);
	m_value->setStat(stat);
	m_value->setSevr(sevr);

	// post a value change event
    // CAS references the event that we sent here, it doesn't copy.
    // If many events are sent in quick succession,
    // (send event before server can broadcast the previous one)
    // it's required to post a copy, otherwise all the events will
    // actually equal the most recent one.
    // For the scalar case, this is done by gdd::Dup,
    // for arrays, it'll only reference the array data.
	if (m_monitored)
	{
		casEventMask    mask(m_server->valueEventMask|m_server->logEventMask);
        if (m_value->isScalar())
        {
            gdd *event = new gdd(m_value->applicationType(), m_value->primitiveType());
            if (m_value->primitiveType() == aitEnumFixedString)
            {   
                aitFixedString s;
                m_value->get(s);
                event->put(s);
            }
            else
                event->copy(m_value);
            event->setTimeStamp(&time);
	        event->setStat(stat);
	        event->setSevr(sevr);
		    postEvent(mask, *event);
            event->unreference(); // ref'ef by CAS only
        }
        else
		    postEvent(mask, *m_value);
		m_changed = FALSE;
	}
}

// Set scalar (long, double) value to be served.
#define Comment(c)
#define ImplementSetScalar(MethodName,aitName,aitEnumName,nativeType)	\
void CEpicsPV::MethodName (nativeType value, struct timespec &time)		\
{																\
	double old_value, change;									\
																\
	TRACE4( "%s(%s) = %g ",										\
		#MethodName, getName(), (double) value);				\
																\
	Comment("Check that we have a gdd and that it's a scalar:")	\
	if (m_value)												\
	{															\
		if (!m_value->isScalar())								\
		{														\
			Comment("replace with scalar")						\
			m_value->unreference();								\
			m_value = new gddScalar(m_value_app, aitEnumName);	\
		}														\
	}															\
	else														\
		m_value = new gddScalar(m_value_app, aitEnumName);		\
																\
	if (m_value)												\
	{															\
		Comment("Deadband test")								\
		if (m_have_previous_value)								\
		{														\
			m_value->get (old_value);							\
			change = value - old_value;							\
			if (change < 0)										\
				change = - change;								\
			m_changed = change >= m_deadband;					\
			TRACE4("(%c %g deadband)\n",						\
				m_changed ? '>' : '<', m_deadband);				\
			if (! m_changed)									\
				return;											\
		}														\
																\
		m_value->put((aitName)value);							\
		m_have_previous_value = TRUE;							\
																\
		aitUint16 stat = epicsAlarmNone;						\
		aitUint16 sevr = epicsSevNone;							\
																\
		if (value <= m_alarm_low)								\
		{														\
			sevr = epicsSevMajor;								\
			stat = epicsAlarmLoLo;								\
		}														\
		else													\
		if (value <= m_alarm_low_warning)						\
		{														\
			sevr = epicsSevMinor;								\
			stat = epicsAlarmLow;								\
		}														\
		else													\
		if (value >= m_alarm_high)								\
		{														\
			sevr = epicsSevMajor;								\
			stat = epicsAlarmHiHi;								\
		}														\
		else													\
		if (value >= m_alarm_high_warning)						\
		{														\
			sevr = epicsSevMinor;								\
			stat = epicsAlarmHigh;								\
		}														\
																\
		UpdateValue(stat, sevr, time);							\
	}															\
}

ImplementSetScalar(SetLongValue,aitInt32,aitEnumInt32,long)
ImplementSetScalar(SetULongValue,aitUint32,aitEnumUint32,unsigned long)
ImplementSetScalar(SetDoubleValue,aitFloat64,aitEnumFloat64,double)

#undef ImplementSetScalar
#undef Comment

// Similar to SetLongValue,
// but checks if value is allowed for this enum
BOOL CEpicsPV::SetEnumValue(long value, struct timespec &time)
{
	TRACE4( "SetEnumValue(%s) = %ld\n", getName(), value);
	// Set the value anyway:
	SetLongValue (value, time);

	return value >= 0  &&  value < (long)m_enum_strings.size();
}

// Strings should be handled by aitString,
// not aitFixedString,
// but I yet have to understand how to
// use that and assert that there
// are no memory problems....
void CEpicsPV::SetStringValue(const char *txt, size_t len, struct timespec &time)
{
	TRACE4( "SetStringValue(%s) = %s\n",getName(), txt);

	// Check that we have a gdd and that it's a scalar:
	if (m_value)
	{
		if (!m_value->isScalar())
		{
			// replace with scalar
			m_value->unreference();
			m_value = new gdd(m_value_app, aitEnumFixedString);
		}
	}
	else
		m_value = new gdd(m_value_app, aitEnumFixedString);

	if (m_value)
	{
		aitFixedString value;
		strncpy(value.fixed_string, txt, AIT_FIXED_STRING_SIZE);
		m_value->put(value);
		UpdateValue(epicsAlarmNone, epicsSevNone, time);
	}
}

// To be used as gddDestructor for linear (1D) arrays
// of type T:
template<class T>
class gddArrayDestructor : public gddDestructor
{
public:
	void run (void *v)
	{
		T *arr = (T *) v;
		delete [] arr;
		TRACE4("gddArrayDestructor\n");
	}
};

// Set array as value to be served.
// This array is copied into the CEpicsPV,
// so the argument passed might be temporary
// (as it is for a COM-retrieved VARIANT).
#define Comment(c)
#define ImplementSetArray(MethodName,aitEnumName,aitName,nativeType)	\
void CEpicsPV::MethodName(unsigned long size, nativeType *data, struct timespec &time)	\
{																		\
	aitIndex start, nelem;												\
																		\
	Comment("Check that we have a matching array gdd")					\
	if (m_value)														\
	{																	\
		if (m_value->dimension() != 1               ||					\
			m_value->getBound(0, start, nelem) != 0 ||					\
			start != 0 || nelem != size)								\
		{	Comment("wrong gdd, replace")								\
			m_value->unreference();										\
			m_value = new gddArray(m_value_app, aitEnumName, 1, size);	\
		}																\
	}																	\
	else																\
	{																	\
		Comment("get new one")											\
		m_value = new gddArray(m_value_app, aitEnumName, 1, size);		\
	}																	\
																		\
	if (m_value)														\
	{																	\
		Comment("This might fail with future architectures:")			\
		ATLASSERT(sizeof(aitName) == sizeof(nativeType));				\
																		\
		Comment("copy array data:")										\
		aitName	*copy = new aitName[size];								\
		if (!copy)														\
			return;														\
		memcpy (copy, data, sizeof(nativeType)*size);					\
		m_value->putRef(copy, new gddArrayDestructor<aitName>);			\
																		\
		Comment("Check deadband for all elements? Now: update always")	\
		m_changed = TRUE;												\
		UpdateValue(epicsAlarmNone, epicsSevNone, time);                \
	}																	\
}

ImplementSetArray(SetShortArray,aitEnumInt16,aitInt16,short)
ImplementSetArray(SetUShortArray,aitEnumUint16,aitUint16,unsigned short)
ImplementSetArray(SetLongArray,aitEnumInt32,aitInt32,long)
ImplementSetArray(SetULongArray,aitEnumUint32,aitUint32,unsigned long)
ImplementSetArray(SetFloatArray,aitEnumFloat32,aitFloat32,float)
ImplementSetArray(SetDoubleArray,aitEnumFloat64,aitFloat64,double)
#undef Comment
#undef ImplementSetArray

// read
//
caStatus CEpicsPV::read (const casCtx &ctx, gdd &value)
{
#if TRACELEVEL >= 3
	unsigned apptype = value.applicationType();
	ATLTRACE ("CEpicsPV::read (%s): requested is '%s' (%d), native type '%s'\n",
		getName(),
		gddApplicationTypeTable::app_table.getName (apptype), apptype,
		aitName[value.primitiveType()]
		);

	if (value.isContainer())
	{
		gddCursor curs = ((gddContainer *)&value)->getCursor();
		gdd *item;

		for (item=curs.first(); item; item=curs.next())
		{
			apptype = item->applicationType();
			ATLTRACE ("\t'%s' (%d), native type '%s'\n",
				gddApplicationTypeTable::app_table.getName (apptype),
				apptype,
				aitName[item->primitiveType()]
		);
		}
	}
#endif
	return m_func_table.read (*this, value);
}


//
// write 
//
// The request is allowed to complete asynchronously
// (see Asynchronous IO Classes below).
//
// Return S_casApp_postponeAsyncIO if too many simultaneous
// asynchronous IO operations are pending aginst the PV. 
// The server library will retry the request whenever an
// asynchronous IO operation (read or write) completes
// against the PV.
//
caStatus CEpicsPV::write (const casCtx &ctx, gdd &value)
{
	CComVariant		*variant = NULL;
	aitString		sval;
	aitFixedString	fsval;
	aitInt32		lval;
	aitFloat64		dval;

	ATLTRACE (_T("CEpicsPV::write (%s)\n"), getName());
#if TRACELEVEL >= 3
	unsigned apptype = value.applicationType();
	ATLTRACE ("CEpicsPV::write (%s): provided is '%s' (%d), native type '%s'\n",
		getName(),
		gddApplicationTypeTable::app_table.getName (apptype), apptype,
		aitName[value.primitiveType()]
		);

	if (value.isContainer())
	{
		gddCursor curs = ((gddContainer *)&value)->getCursor();
		gdd *item;

		for (item=curs.first(); item; item=curs.next())
		{
			apptype = item->applicationType();
			ATLTRACE ("\t'%s' (%d), native type '%s'\n",
				gddApplicationTypeTable::app_table.getName (apptype),
				apptype,
				aitName[item->primitiveType()]
		);
		}
	}
#endif
	if (m_pv == NULL || value.isContainer())
		return S_casApp_noSupport;

	if (value.isScalar())
	{
		switch (value.primitiveType())
		{	// there is no 'unsigned' subtype in VARIANT...
		case aitEnumInt8:
		case aitEnumUint8:
		case aitEnumInt16:
		case aitEnumUint16:
		case aitEnumEnum16:
		case aitEnumInt32:
		case aitEnumUint32:
			value.get (lval);
			variant = new CComVariant ((long) lval);
			break;
		case aitEnumFloat32:
		case aitEnumFloat64:
			value.get (dval);
			variant = new CComVariant ((double) dval);
			break;
		case aitEnumString:
			value.get (sval);
			variant = new CComVariant (sval.string());
			break;
		case aitEnumFixedString:
			value.get (fsval);
			variant = new CComVariant (fsval.fixed_string);
			break;
		default:
			return S_casApp_noSupport;
		}
	}
	else
	{	// Array
		aitIndex start, nelem;
		if (value.dimension() != 1   ||
			value.getBound(0, start, nelem) != 0)
		{
			ATLTRACE ("EpicsPV::write: cannot handle multidim. array\n");
			return S_casApp_noSupport;
		}
		VARTYPE  vt;
		size_t	size;

		switch (value.primitiveType())
		{
		case aitEnumFloat32:
			vt = VT_R4;
			size=sizeof(float);
			break;
		case aitEnumFloat64:
			vt = VT_R8;
 			size=sizeof(double);
			break;
		case aitEnumInt8:
		case aitEnumUint8:
			vt = VT_UI1;
			size = 1;
			break;
		case aitEnumInt16:
		case aitEnumUint16:
		case aitEnumEnum16:
			vt = VT_I2;
			size=2;
			break;
		case aitEnumInt32:
		case aitEnumUint32:
			vt = VT_I4;
			size = 4;
			break;
		default:
			return S_casApp_noSupport;
		}
		SAFEARRAYBOUND rgsabound;
		rgsabound.lLbound = start;
		rgsabound.cElements = nelem;
		SAFEARRAY *arr = SafeArrayCreate (vt, 1, &rgsabound);
		if (!arr)
			return S_casApp_noSupport;

		void *arrdata;
		SafeArrayAccessData (arr, &arrdata);
 		memcpy (arrdata, value.dataVoid(), nelem*size);
		SafeArrayUnaccessData (arr);
		variant = new CComVariant ();
		variant->parray = arr;
		variant->vt = VT_ARRAY | vt;
	}

	if (variant)
	{
		ATLTRACE ("Fire_Changed\n");
		m_pv->Fire_Changed(*variant);
		delete variant;
		return S_cas_success;
	}
	
	return S_casApp_noSupport;
}

// Actual read routines for various app-types,
// all called via m_func_table
gddAppFuncTableStatus CEpicsPV::getUnits(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getUnits (%s)\n"), getName());
	value.put (m_units.c_str());

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getAlarmLowWarning(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getAlarmLowWarning (%s)\n"), getName());
	value.put (m_alarm_low_warning);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getAlarmHighWarning(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getAlarmHighWarning (%s)\n"), getName());
	value.put (m_alarm_high_warning);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getAlarmLow(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getAlarmLow (%s)\n"), getName());
	value.put (m_alarm_low);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getAlarmHigh(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getAlarmHigh (%s)\n"), getName());
	value.put (m_alarm_high);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getGraphicLow(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getGraphicLow (%s)\n"), getName());
	value.put (m_graphic_low);

	return S_casApp_success;
}


gddAppFuncTableStatus CEpicsPV::getGraphicHigh(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getGraphicHigh (%s)\n"), getName());
	value.put (m_graphic_high);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getControlLow(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getControlLow (%s)\n"), getName());
	value.put (m_control_low);

	return S_casApp_success;
}


gddAppFuncTableStatus CEpicsPV::getControlHigh(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getControlHigh (%s)\n"), getName());
	value.put (m_control_high);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getPrecision(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getPrecision (%s)\n"), getName());
	value.put (m_precision);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getEnums(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getEnums (%s)\n"), getName());
	unsigned		nStr = m_enum_strings.size ();

	if (nStr < 1)
	{
		// ?? This seems to be no enum:
		// pass current value ?
		return S_casApp_undefined;
	}

	aitFixedString	*str = new aitFixedString[nStr];
	if (!str)
		return S_casApp_noMemory;

	for (unsigned i=0; i<nStr; ++i)
	{
		strncpy (str[i].fixed_string,
			m_enum_strings[i].c_str(),
			AIT_FIXED_STRING_SIZE);
			// = sizeof(str[i].fixed_string));
	}

	value.setDimension(1);
	value.putRef (str, new gddArrayDestructor<aitFixedString>);
	value.setBound (0,0,nStr);

	return S_casApp_success;
}

gddAppFuncTableStatus CEpicsPV::getValue(gdd &value)
{
	TRACE4 (_T("CEpicsPV::getValue (%s)\n"), getName());

	if (m_value == NULL)
		return S_casApp_undefined;

	value.put (m_value);

	return S_casApp_success;
}

// Called by caServer several times,
// only effective when we are no longer registered
// with CServer:
void CEpicsPV::destroy ()
{
	if (! m_keep_registered)
	{
		m_server->Unregister (*this);
		TRACE2 (_T("CEpicsPV::destroy ('%s') - Harakiri\n"), getName());
		casPV::destroy ();
		return;
	}
	TRACE2 (_T("CEpicsPV::destroy ('%s') - NOP\n"), getName());
}


// EOF EpicsPV.cpp

