// EpicsPV.h: interface for the CEpicsPV class.
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EPICSPV_H__0E17D828_E06E_11D2_8F16_00105AC8D47C__INCLUDED_)
#define AFX_EPICSPV_H__0E17D828_E06E_11D2_8F16_00105AC8D47C__INCLUDED_

#include <casdef.h>
#include <gddAppFuncTable.h>
#include <string>
#include <vector>

using std::string;
using std::vector;

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//	CEpicsPV is the casPV-derived class for this server,
//	CPV handles the COM interface.
//	Both are linked and can only exist as twins,
//	still they are separated because it's hopefully
//	easier to see the COM/CA differences:
//
//	I prefer uppercase method names,
//	so this makes it easy to spot the lowercase names
//	which refer to virtual overrides required by casPV.
class CServer;
class CProcessVariable;
class CEpicsPV :
	public casPV,				// has to be a casPV, of course
	public tsSLNode<CEpicsPV>,	// for hash table: needs to be put in list
	public stringId				// also needed for hash table, is set to name
{
public:
	// Static initialization,
	// automatically called when CServer is created:
	static void StaticInit (CServer *server);

	// constructor cannot return status,
	// so this Create-routine should be used to obtain
	// a new object.
	// Create also handles registration with server:
	static CEpicsPV *Create(const char *name, CProcessVariable *pv);

	virtual ~CEpicsPV();

    // A PV could have been created but no value
    // was assigned, yet -> type is undefined!
    bool isDefined() const
    {   return m_value != 0; }

	// Nomen est omen:
	const char *getName() const;

	// units cannot be NULL, use "" instead !
	const char *GetUnits() const
	{
		return m_units.c_str();
	}

	void SetUnits (const char *units)
	{
		m_units = units;
	}

	// Called by the server libary each time that it wishes to
	// subscribe for PV change notification from the server 
	// tool via postEvent() below.
	caStatus interestRegister ();

	// called by the server library each time that it wishes to
	// remove its subscription for PV value change events
	// from the server tool via caServerPostEvents()
	void interestDelete ();

	// tbe best type for clients to use when accessing the
	// value of the PV
	aitEnum bestExternalType() const;

	// dimension (0: scalar, 1: array)
	unsigned maxDimension() const; // return zero if scaler

	// return array size for given dimension (starting with dimension=0)
	aitIndex maxBound (unsigned dimension) const;

	// Configure PV
	// Post monitor when changing these params?
	void SetAlarmLowWarning (double value)
	{
		m_alarm_low_warning = value;
	}
	void SetAlarmHighWarning (double value)
	{
		m_alarm_high_warning = value;
	}
	void SetAlarmLow (double value)
	{
		m_alarm_low = value;
	}
	void SetAlarmHigh (double value)
	{
		m_alarm_high = value;
	}
	void SetGraphicLow (double value)
	{
		m_graphic_low = value;
	}
	void SetGraphicHigh (double value)
	{
		m_graphic_high = value;
	}
	void SetPrecision (int value)
	{
		m_precision = value;
	}
	void SetControlLow (double value)
	{
		m_control_low = value;
	}
	void SetControlHigh (double value)
	{
		m_control_high = value;
	}
	void SetDeadband (double value)
	{
		m_deadband = value;
	}

	double GetAlarmLowWarning ()
	{
		return m_alarm_low_warning;
	}
	double GetAlarmHighWarning ()
	{
		return m_alarm_high_warning;
	}
	double GetAlarmLow ()
	{
		return m_alarm_low;
	}
	double GetAlarmHigh ()
	{
		return m_alarm_high;
	}
	double GetGraphicLow ()
	{
		return m_graphic_low;
	}
	double GetGraphicHigh ()
	{
		return m_graphic_high;
	}
	int GetPrecision ()
	{
		return m_precision;
	}
	double GetControlLow ()
	{
		return m_control_low;
	}
	double GetControlHigh ()
	{
		return m_control_high;
	}
	double GetDeadband ()
	{
		return m_deadband;
	}

	BOOL SetEnumString (int index, const char *text);

    static void getEpicsTime(struct timespec &now);
	
	// Set long or double value to be served.
    void SetLongValue (long value, struct timespec &time);
	void SetULongValue (unsigned long value, struct timespec &time);
	void SetDoubleValue (double value, struct timespec &time);
	void SetStringValue (const char *txt, size_t len, struct timespec &time);
	// Similar, but checks if value is allowed for this enum
	BOOL SetEnumValue (long value, struct timespec &time);

	// Set double array as value to be served.
	// This array is copied into the CEpicsPV,
	// so the argument passed might be temporary
	// (as it is for a COM-retrieved VARIANT).
	void SetShortArray (unsigned long size, short *data, struct timespec &time);
	void SetUShortArray (unsigned long size, unsigned short *data, struct timespec &time);
	void SetLongArray (unsigned long size, long *data, struct timespec &time);
	void SetULongArray (unsigned long size, unsigned long *data, struct timespec &time);
	void SetFloatArray (unsigned long size, float *data, struct timespec &time);
	void SetDoubleArray (unsigned long size, double *data, struct timespec &time);

	// read
	caStatus read (const casCtx &ctx, gdd &prototype);

	// write 
	caStatus write (const casCtx &ctx, gdd &value);

	// CServer will call this when it's going down
	// and intends to remove all PVs.
	// CPV (IPV) will call this when no COM client
	// is interested on providing this PV any more.
	void Unregister ()
	{
		m_keep_registered = false;
	}

	// Called by caServer several times,
	// only effective when we are no longer registered
	// with CServer:
	void destroy ();

private:
	// stringId has to be initialized,
	// but constructor is private because Create should be used:
	CEpicsPV(const char *name, CProcessVariable *pv);

	CProcessVariable *m_pv;

	// These are the data that is actually held
	// by this PV:
	static aitUint32 m_value_app;
	gdd *m_value;	// the numeric value
	BOOL m_have_previous_value;
	string	m_units;
	
	// support for enumerated types:
	// (assume we need no enums, but adjust by 10 if so)
	vector<string>	m_enum_strings;

	// only value changes exceeding deadband will be 'published'
	double m_deadband; 
	BOOL   m_changed;

	double m_alarm_low;
	double m_alarm_low_warning;
	double m_alarm_high_warning;
	double m_alarm_high;
	double m_graphic_low;
	double m_graphic_high;
	int	m_precision;
	double m_control_low;
	double m_control_high;

	// is this PV monitored?
	BOOL m_monitored;

	// Called after the "value" of m_value is set
	// to update the time stamp, status and severity
    void UpdateValue(aitUint16 stat, aitUint16 sevr, struct timespec &time);

	// As long as this PV is registered by the server,
	// destroy() must not delete it:
	BOOL m_keep_registered;

	// pointer to the one-and-only CServer
	// where this PV is registered:
	static CServer	*m_server;

	// Dispatch-table for read-routines:
	static gddAppFuncTable<CEpicsPV>	m_func_table;

	// Actual read routines for various app-types,
	// all called via m_func_table
	gddAppFuncTableStatus getUnits(gdd &value);
	gddAppFuncTableStatus getAlarmLowWarning(gdd &value);
	gddAppFuncTableStatus getAlarmHighWarning(gdd &value);
	gddAppFuncTableStatus getAlarmLow(gdd &value);
	gddAppFuncTableStatus getAlarmHigh(gdd &value);
	gddAppFuncTableStatus getGraphicLow(gdd &value);
	gddAppFuncTableStatus getGraphicHigh(gdd &value);
	gddAppFuncTableStatus getPrecision(gdd &value);
	gddAppFuncTableStatus getValue(gdd &value);
	gddAppFuncTableStatus getControlLow(gdd &value);
	gddAppFuncTableStatus getControlHigh(gdd &value);
	gddAppFuncTableStatus getEnums(gdd &value);
};

#endif // !defined(AFX_EPICSPV_H__0E17D828_E06E_11D2_8F16_00105AC8D47C__INCLUDED_)
