// ProcessVariable.cpp : Implementation of CProcessVariable
#include "stdafx.h"
#include "Config.h"
#include "EpicsCAServer.h"
#include "ProcessVariable.h"

#include "Server.h"
#include "EpicsPV.h"
#include "ArrayTools.h"
#ifdef _DEBUG
#include "VariantTools.h"
#endif

#ifdef _DEBUG
SLList<CProcessVariable>	debug_PV_list;
#endif

extern CServer *the_server;

/////////////////////////////////////////////////////////////////////////////
// CProcessVariable

CProcessVariable::CProcessVariable()
{
	m_epics_pv = NULL;
#ifdef _DEBUG
	debug_PV_list.add (this);
#endif
}

CProcessVariable::~CProcessVariable()
{
	ATLTRACE (_T("CProcessVariable::~CProcessVariable: IPV for '%s' removed\n"),
		m_epics_pv ? m_epics_pv->getName() : "<unnamed>");
#ifdef _DEBUG
	debug_PV_list.remove (this);
#endif
	if (m_epics_pv != NULL)
	{
		m_epics_pv->Unregister ();
		m_epics_pv->destroy ();
		m_epics_pv = NULL;
	}
}

STDMETHODIMP CProcessVariable::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IProcessVariable
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

STDMETHODIMP CProcessVariable::put_name(BSTR newVal)
{
	USES_CONVERSION;
	LPCSTR	name;
	size_t len;

	ATLTRACE (_T("CProcessVariable::put_name('%hs')\n"), OLE2CA(newVal));

	//	The corresponding EpicsPV cannot change it't name:
	if (m_epics_pv != NULL)
	{
		Error ("Once set, Process Variable name cannot be modified", IID_IProcessVariable);
		return E_FAIL;
	}
	
	name = OLE2CA(newVal);
	len = strlen(name);
	if (len <= 0)
	{
		Error ("Process Variable name cannot have zero-length", IID_IProcessVariable);
		return E_FAIL;
	}

	if (the_server)
    {
        if (the_server->findPV (name))
	    {
		    Error ("Process Variable already exists on this server", IID_IProcessVariable);
		    return E_FAIL;
	    }
    }

	// Create Epics PV, PV registers with CServer etc.:
	m_epics_pv = CEpicsPV::Create (name, this);
	if (! m_epics_pv)
	{
		Error ("EpicsCAServer: internal error, cannot setup PV",
			IID_IProcessVariable);
		return E_FAIL;
	}
	
	return S_OK;
}

STDMETHODIMP CProcessVariable::get_name(BSTR *pVal)
{
	if (m_epics_pv)
	{
		CComBSTR	name (m_epics_pv->getName());
		*pVal = name;
	}
	else
	{
		CComBSTR	name ("<uninitialized>");
		*pVal = name;
	}

	return S_OK;
}


STDMETHODIMP CProcessVariable::get_units(BSTR *pVal)
{
	if (m_epics_pv)
	{
		CComBSTR	units (m_epics_pv->GetUnits());
		*pVal = units;
	}
	else
	{
		CComBSTR	units ("<uninitialized>");
		*pVal = units;
	}

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_units(BSTR newVal)
{
	if (m_epics_pv == NULL)
	{
		Error ("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}
	
	USES_CONVERSION;
	LPCSTR	units = OLE2CA(newVal);

	ATLTRACE (_T("CPV::put_units('%hs')\n"), units);
	
	m_epics_pv->SetUnits(units);
	
	return S_OK;
}

STDMETHODIMP CProcessVariable::get_low_warning(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetAlarmLowWarning();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_low_warning(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetAlarmLowWarning(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_high_warning(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetAlarmHighWarning();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_high_warning(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetAlarmHighWarning(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_alarm_low(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetAlarmLow();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_alarm_low(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetAlarmLow(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_alarm_high(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetAlarmHigh();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_alarm_high(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetAlarmHigh(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_graphic_low(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetGraphicLow();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_graphic_low(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetGraphicLow(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_graphic_high(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetGraphicHigh();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_graphic_high(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetGraphicHigh(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_control_low(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetControlLow();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_control_low(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetControlLow(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_control_high(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetControlHigh();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_control_high(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetControlHigh(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_precision(long *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetPrecision();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_precision(long newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetPrecision(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::get_deadband(double *pVal)
{
	if (m_epics_pv)
		*pVal = m_epics_pv->GetDeadband();

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_deadband(double newVal)
{
	if (m_epics_pv)
		m_epics_pv->SetDeadband(newVal);

	return S_OK;
}

STDMETHODIMP CProcessVariable::put_enum_string(int index, BSTR newVal)
{
	USES_CONVERSION;

	ATLTRACE ("Enum String #%d: %s\n", index, OLE2CA(newVal));
	if (!m_epics_pv)
	{
		Error ("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}

	if (! m_epics_pv->SetEnumString (index, OLE2CA(newVal)))
	{
		Error ("Cannot set enum string, check index", IID_IProcessVariable);
		return E_FAIL;
	}

	return S_OK;
}

STDMETHODIMP CProcessVariable::SetValue(VARIANT new_value)
{
    struct timespec now;
    CEpicsPV::getEpicsTime(now);
    return SetValue(new_value, now);
}

STDMETHODIMP CProcessVariable::SetValueAndTime(VARIANT new_value, unsigned long secs, unsigned long nanosecs)
{
    struct timespec time;
    time.tv_sec = secs;
    time.tv_nsec = (long) nanosecs;
    return SetValue(new_value, time);
}

STDMETHODIMP CProcessVariable::SetEnumValue(long value)
{
    struct timespec now;
    CEpicsPV::getEpicsTime(now);
    return SetEnumValueAndTime(value, (unsigned long)now.tv_sec, (unsigned long)now.tv_nsec);
}

STDMETHODIMP CProcessVariable::SetEnumValueAndTime(long value, unsigned long secs, unsigned long nanosecs)
{
	if (m_epics_pv == NULL)
	{
		Error ("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}
    struct timespec time;
    time.tv_sec = secs;
    time.tv_nsec = (long) nanosecs;
	if (! m_epics_pv->SetEnumValue (value, time))
	{
		Error ("Invalid value for enumerated PV", IID_IProcessVariable);
		return E_FAIL;
	}

	return S_OK;
}

static void BSTR2str (const OLECHAR *bstr, char *buf, size_t &len)
{
	if (! bstr)
	{
		len = 0;
		*buf = '\0';
		return;
	}

	char *dest = buf;
	size_t l = SysStringLen ((BSTR) bstr);
	if (len <= l)
		l = len-1;
	len = l;
	while (l > 0)
	{
		*(buf++) = *(bstr++);
		--l;
	}
	*(buf++) = '\0';
}

HRESULT CProcessVariable::SetValue(VARIANT new_value, struct timespec &time)
{
	if (m_epics_pv == NULL)
	{
		Error("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}

	VARTYPE type = new_value.vt & VT_TYPEMASK;
#if TRACELEVEL >= 4
	ATLTRACE(_T("CProcessVariable::SetValue (%s), VARIANT type %s\n"),
		m_epics_pv->getName(), VariantTypeString(new_value));
#endif

	// handle arrays in sep. routine:
	if (new_value.vt & VT_ARRAY)
		return SetArrayValue(type, new_value.parray, time);

	// dispatch VARIANT type for scalars
	switch (type)
	{
	case VT_I2:
		m_epics_pv->SetLongValue((long) new_value.iVal, time);
		return S_OK;
    case VT_UI2:
		m_epics_pv->SetULongValue((unsigned long) new_value.uiVal, time);
		return S_OK;
	case VT_I4:
		m_epics_pv->SetLongValue(new_value.lVal, time);
		return S_OK;
    case VT_UI4:
		m_epics_pv->SetULongValue(new_value.ulVal, time);
		return S_OK;
	case VT_R4:
		m_epics_pv->SetDoubleValue((double) new_value.fltVal, time);
		return S_OK;
	case VT_R8:
		m_epics_pv->SetDoubleValue(new_value.dblVal, time);
		return S_OK;
	case VT_BSTR:
		char text[AIT_FIXED_STRING_SIZE];
		size_t len = AIT_FIXED_STRING_SIZE;
		BSTR2str(new_value.bstrVal, text, len);
		m_epics_pv->SetStringValue(text, len, time);
		return S_OK;
	}

	Error("Can only handle: short,int,long,float,double, BSTR", IID_IProcessVariable);
	return E_FAIL;
}

HRESULT CProcessVariable::SetArrayValue(VARTYPE type, SAFEARRAY *arr, struct timespec &time)
{
	UINT dim = SafeArrayGetDim(arr);
	unsigned long size;
	void HUGEP *data;

	if (! ( (dim == 1)  ||
            (dim == 2  && arr->rgsabound[1].cElements == 1)
            ))
	{
		Error("Can only handle single-dimensional arrays", IID_IProcessVariable);
		return E_FAIL;
	}
    
	//	'1' for first dimension, not '0' !
	size = arr->rgsabound[0].cElements;
#if TRACELEVEL >= 4
	ATLTRACE(_T("%d elements\n"), size);
#endif

	SafeArrayAccessData(arr, &data);
	switch (type)
	{
	case VT_I2:
		m_epics_pv->SetShortArray(size, (short *)data, time);
		break;
	case VT_UI2:
		m_epics_pv->SetUShortArray(size, (unsigned short *)data, time);
		break;
	case VT_I4:
		m_epics_pv->SetLongArray(size, (long *)data, time);
		break;
	case VT_UI4:
		m_epics_pv->SetULongArray(size, (unsigned long *)data, time);
		break;
	case VT_R4:
		m_epics_pv->SetFloatArray(size, (float *)data, time);
		break;
	case VT_R8:
		m_epics_pv->SetDoubleArray(size, (double *)data, time);
		break;
	default:
		Error("Can only handle arrays of short,int,long,float,double. "
              "Cannot handle VARIANT arrays of VARIANTS.", IID_IProcessVariable);
		SafeArrayUnaccessData(arr);
		return E_FAIL;
	}
	SafeArrayUnaccessData(arr);

	return S_OK;
}

STDMETHODIMP CProcessVariable::GetEpicsTime(unsigned long *seconds, unsigned long *nanosecs)
{
    if (seconds && nanosecs)
    {
        struct timespec now;
        CEpicsPV::getEpicsTime(now);
        *seconds = (unsigned long)now.tv_sec;
        *nanosecs= (unsigned long)now.tv_nsec;
    }

	return S_OK;
}

STDMETHODIMP CProcessVariable::SetValuesAndTimes(VARIANT values, VARIANT times)
{
	if (m_epics_pv == NULL)
	{
		Error("PV name not set", IID_IProcessVariable);
		return E_FAIL;
	}

	if (!(values.vt & VT_ARRAY))
	{
		Error("Expect array of values", IID_IProcessVariable);
		return E_FAIL;
	}
	if (times.vt != (VT_ARRAY|VT_UI4))
	{
		Error("Expect array of UI4/U32 for time stamps", IID_IProcessVariable);
		return E_FAIL;
	}

	if (SafeArrayGetDim(values.parray) != 1)
	{
		Error("Can only handle single-dimensional values array", IID_IProcessVariable);
		return E_FAIL;
	}
	long lower, upper;
	SafeArrayGetLBound(values.parray, 1, &lower);
	SafeArrayGetUBound(values.parray, 1, &upper);
	unsigned long value_size = upper - lower + 1; // lower..upper -> 0...size-1

	if (SafeArrayGetDim(times.parray) != 1)
	{
		Error("Can only handle single-dimensional times array", IID_IProcessVariable);
		return E_FAIL;
	}
	SafeArrayGetLBound(times.parray, 1, &lower);
	SafeArrayGetUBound(times.parray, 1, &upper);
	unsigned long time_size = upper - lower + 1;

    if (2*value_size != time_size)
	{
		Error("Array size error, expect two time entries [sec, nano] per value", IID_IProcessVariable);
		return E_FAIL;
	}

	void *value_data;
    unsigned long *time_data;
	SafeArrayAccessData(values.parray, &value_data);
	SafeArrayAccessData(times.parray, (void **)&time_data);
    struct timespec time;
    for (unsigned long i=0; i<value_size; ++i)
    {
        time.tv_sec = *(time_data++);
        time.tv_nsec = (long) *(time_data++);
        switch (values.vt & VT_TYPEMASK)
        {
    	case VT_I2:
	    	m_epics_pv->SetLongValue(*((short *)value_data), time);
            value_data = ((char *)value_data) + 2;
            break;
    	case VT_UI2:
	    	m_epics_pv->SetULongValue(*((unsigned short *)value_data), time);
            value_data = ((char *)value_data) + 2;
            break;
    	case VT_I4:
	    	m_epics_pv->SetLongValue(*((long *)value_data), time);
            value_data = ((char *)value_data) + 4;
            break;
    	case VT_UI4:
	    	m_epics_pv->SetULongValue(*((float *)value_data), time);
            value_data = ((char *)value_data) + sizeof(float);
            break;
    	case VT_R4:
	    	m_epics_pv->SetDoubleValue(*((double *)value_data), time);
            value_data = ((char *)value_data) + sizeof(double);
            break;
    	case VT_R8:
	    	m_epics_pv->SetDoubleValue(*((double *)value_data), time);
            value_data = ((char *)value_data) + sizeof(double);
            break;
    	default:
		    SafeArrayUnaccessData(times.parray);
		    SafeArrayUnaccessData(values.parray);
	    	Error("Can only handle values of double", IID_IProcessVariable);
		    return E_FAIL;
	    }
    }
    SafeArrayUnaccessData(times.parray);
	SafeArrayUnaccessData(values.parray);

	return S_OK;
}
